import os
import argparse
from tqdm import tqdm
from validate import *
from texts import *
from output import *


def process(filename: str, output_folder: str, output_format: str):
    validate_file_path(filename)
    validate_output_folder(output_folder)
    validate_output_format(output_format)

    readdata = False
    output_path = os.path.join(output_folder, f"{os.path.splitext(
        os.path.basename(filename))[0]}.{output_format}")

    with open(filename, 'r') as infile:
        schema = {"relation": "", "attributes": [], "data": []}

        for line in infile:
            if line[0] == "%":
                continue
            elif line[0] == "@":
                args = line.split()
                if args[0] == "@relation":
                    schema["relation"] = args[1]
                elif args[0] == "@attribute":
                    values = "".join(args[2:])
                    if values[0] == "{":
                        values = values[1:len(values)-1]
                        values = values.split(",")
                    schema["attributes"].append(
                        {"name": args[1], "values": values})
                elif args[0] == "@data":
                    readdata = True
            elif readdata:
                schema["data"].append(line.strip())

        names = []
        data = []
        attrs = schema["attributes"]

        for attr in attrs:
            names.append(attr["name"])

        for row in tqdm(schema["data"], desc="Reading From File", unit="row", dynamic_ncols=True, leave=False, colour="blue"):
            row = row.split(",")
            entry = {}
            for name in names:
                entry[name] = row[names.index(name)]
            data.append(entry)

        schema["data"] = data

        if output_format == "xml" or output_format == "xlsx" or output_format == "orc":
            build_output(schema, output_path, output_format)
        else:
            with open(output_path, 'w') as outfile:
                build_output(schema, outfile, output_format)

    print(f"\nFile converted into {output_format} format.")
    print(f"Output saved to: {os.path.abspath(output_path)}")
    print("Thanks for using arff-format-converter")


def main():
    parser = argparse.ArgumentParser(
        description="Convert ARFF files to different formats.",
        epilog=bottom_msg,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "--file", "-f", help="Path to the ARFF file.", type=str, required=True)
    parser.add_argument(
        "--output", "-o", help="Path to the output folder.", type=str, required=True)
    parser.add_argument(
        "--format", "-fmt", help="Output format: 'xml', 'json', 'csv', 'xlsx', 'orc'.",
        choices=["xml", "json", "csv", "xlsx", "orc"],
        type=str, required=True)

    args = parser.parse_args()

    process(args.file, args.output, args.format)



from logs import *
import os


def validate_file_path(path: str):
    if not os.path.isfile(path):
        log_error(
            0, FILE_ERROR, "Invalid file path. Please provide a valid path to the ARFF file.")
        exit_with_errors()


def validate_output_folder(path: str):
    if not os.path.isdir(path):
        log_error(
            0, FILE_ERROR, "Invalid output folder path. Please provide a valid path to the output folder.")
        exit_with_errors()


def validate_output_format(output_format: str):
    if output_format not in legal_formats:
        log_error(0, OPTION_ERROR,
                  "Invalid output format. Please provide a valid output format.")
        exit_with_errors()



import ujson as json
import csv
import fastavro
import pandas as pd
from validate import *
from tqdm import tqdm


try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET


# XML Encodeing Start
def build_nodes(data, name, parent):
    if isinstance(data, list):
        for item in data:
            build_nodes(item, name, ET.SubElement(parent, name))
    elif isinstance(data, dict):
        for attr in data:
            build_nodes(data[attr], name=attr, parent=parent)
    else:
        parent.text = data


def output_xml(schema, outfile):
    progress_bar = tqdm(total=100, desc='Writing XML', unit=" %",
                        position=5, leave=False, colour="red")
    progress_bar.update(25 - progress_bar.n)
    root = ET.Element('dataset')
    progress_bar.update(50 - progress_bar.n)
    build_nodes(schema, 'dataset', root)
    progress_bar.update(70 - progress_bar.n)
    ET.ElementTree(root).write(outfile, encoding="utf-8", xml_declaration=True)
    progress_bar.update(100 - progress_bar.n)
# XML Encoding End


# JSON Encoding Start
def output_json(schema, outfile):
    for line in tqdm(json.dumps(schema, ensure_ascii=False, escape_forward_slashes=False),
                     desc="Writing JSON", unit=" words", leave=False, colour="red"):
        outfile.write(line)
# JSON Encoding End


# CSV Encoding Start
def output_csv(schema, outfile):
    progress_bar = tqdm(total=100, desc='Writing CSV', unit=" %",
                        position=5, leave=False, colour="red")
    progress_bar.update(10 - progress_bar.n)
    csv_writer = csv.DictWriter(
        outfile, fieldnames=[attr["name"] for attr in schema["attributes"]])
    progress_bar.update(25 - progress_bar.n)
    csv_writer.writeheader()
    progress_bar.update(50 - progress_bar.n)
    csv_writer.writerows(schema["data"])
    progress_bar.update(100 - progress_bar.n)
# CSV Encoding End


# Excel Encoding Start
def output_excel(schema, outfile):
    chunk_size = 1_000_000  # Set the chunk size as needed
    engine = "openpyxl"

    df = pd.DataFrame(schema["data"])

    # Calculate the number of chunks
    num_chunks = (len(df) // chunk_size) + 1

    with pd.ExcelWriter(outfile, engine=engine) as writer:
        for i in tqdm(range(num_chunks), desc="Writing Excel Sheets", unit=" sheet", dynamic_ncols=True, leave=True, colour="green"):
            start_idx = i * chunk_size
            end_idx = (i + 1) * chunk_size
            chunk_df = df.iloc[start_idx:end_idx, :]

            sheet_name = f'Sheet_{i}'
            chunk_df.to_excel(writer, sheet_name=sheet_name,
                              index=False, engine=engine, header=True)
# Excel Encoding End


# OCR Encoding Start
def output_orc(schema, outfile):
    progress_bar = tqdm(total=100, desc='Writing ORC', unit=" %",
                        position=20, leave=False, colour="red")
    try:
        table = pd.DataFrame(schema["data"])

        avro_schema = {
            "type": "record",
            "name": "RecordName",
            "fields": [
                {"name": col, "type": ["null", "string"]} for col in table.columns
            ]
        }

        with open(outfile, 'wb') as orc_file:
            fastavro.writer(orc_file, schema=avro_schema,
                            records=table.to_dict(orient='records'))
    except ImportError:
        log_error(
            0, OTHER_ERROR, "Fastavro library not installed. Please install it using 'pip install fastavro'.")
    progress_bar.update(100 - progress_bar.n)
# OCR Encoding End


OUTPUT_FORMAT_FUNCTIONS = {
    'json': output_json,
    'csv': output_csv,
    'xlsx': output_excel,
    'orc': output_orc,
    'xml': output_xml
}


def build_output(schema, outfile, output_format: str):
    if output_format in OUTPUT_FORMAT_FUNCTIONS:
        try:
            OUTPUT_FORMAT_FUNCTIONS[output_format](
                schema, outfile)
        except Exception as e:
            log_error(0, OTHER_ERROR, "Error Occured. Please try Again")
            exit_with_errors()
    else:
        log_error(0, OPTION_ERROR,
                  "Invalid output format. Please provide a valid output format.")
        exit_with_errors()



help_message = '''Convert ARFF files to different formats.
The arff-converter tool allows you to convert ARFF files to various output formats. Below are the details:

SYNOPSIS:
    arff-format-converter -f <file> -o <output_folder> -fmt <output_format>

EXAMPLES:
    arff-format-converter -f data.arff -o output -fmt json
    arff-format-converter -f data.arff -o output -fmt xml
    arff-format-converter -f data.arff -o output -fmt csv
    arff-format-converter -f data.arff -o output -fmt xlsx
    arff-format-converter -f data.arff -o output -fmt orc

OPTIONS:
    -f, --file      Path to the ARFF file.
    -o, --output    Path to the output folder.
    -fmt, --format  Output format: 'xml', 'json', 'csv', 'xlsx', 'orc'.

SUPPORTED FORMATS:
    - ARFF (input)
    - XML (output)
    - JSON (output)
    - CSV (output)
    - XLSX (output)
    - ORC (Apache ORC format) (output)

AUTHOR:
    Written by Shani Sinojiya <https://www.shanisinojiya.com>.

REPORTING BUGS:
    Report bugs to <https://github.com/Shani-Sinojiya/arff-format-converter/issues>

COPYRIGHT:
    Copyright © 2024
    License: GPLv3+ - GNU GPL version 3 or later <https://gnu.org/licenses/gpl.html>.
    This is free software: you are free to change and redistribute it.
    There is NO WARRANTY, to the extent permitted by law.
'''

bottom_msg = '''SUPPORTED FORMATS:
    - ARFF (input)
    - XML (output)
    - JSON (output)
    - CSV (output)
    - XLSX (output)
    - ORC (Apache ORC format) (output)

AUTHOR:
    Written by Shani Sinojiya <https://www.shanisinojiya.com>.

REPORTING BUGS:
    Report bugs to <https://github.com/Shani-Sinojiya/arff-format-converter/issues>

COPYRIGHT:
    Copyright © 2024
    License: GPLv3+ - GNU GPL version 3 or later <https://gnu.org/licenses/gpl.html>.
    This is free software: you are free to change and redistribute it.
    There is NO WARRANTY, to the extent permitted by law.'''



from setuptools import setup, find_packages

version = '1.0.3-beta-2'
name = 'arff-format-converter'
description = 'Converts ARFF files to CSV, JSON, XML, XLSX, and ORC'
author = 'Shani Sinojiya'
author_email = 'shanisinojiya@gmail.com'

keywords = [
    "arff",
    "data-conversion",
    "format-conversion",
    "data-interchange",
    "machine-learning",
    "data-preprocessing",
    "data-transformation",
    "file-format-conversion",
    "data-science",
    "python-package",
    "xml",
    "json",
    "csv",
    "excel",
    "orc",
    "pandas",
    "pyarrow",
    "data-manipulation",
    "data-export",
    "data-import",
]

license = open('LICENSE', encoding="utf-8").read()

setup(
    name=name,
    version=version,
    packages=find_packages(),
    description=description,
    author=author,
    author_email=author_email,
    maintainer=author,
    maintainer_email=author_email,
    license=license,
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        "ujson",
        "pandas",
        "argparse",
        "tqdm",
        "fastavro",
        "openpyxl",
    ],
    entry_points={
        'console_scripts': [
            'arff-format-converter=arff_format_converter.arff_converter:main',
        ],
    },
    keywords=keywords,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



import sys

FILE_ERROR = 0
OPTION_ERROR = 1
OTHER_ERROR = 2
legal_formats = ["xml", "json", "csv", "xlsx", "orc"]
error_array = ["Invalid file extension - expects '<filename>.arff'",
               "Invalid option - expects 'xml', 'json', 'csv', 'xlsx', or 'orc'.", "The file format is invalid."]
error_log = []


def console(msg):
    print(msg, file=sys.stderr)


def log_error(line_num, error_index, details):
    this_error = {"line_num": line_num,
                  "message": error_array[error_index], "details": details}
    error_log.append(this_error)
    console(f"Error at {line_num}:")
    console(error_array[error_index])
    console(details)


def exit_with_errors():
    console("Exiting with errors.")
    sys.exit(1)
